#define INCL_PM
#define INCL_DOSPROCESS

#include <string.h>
#include <stdio.h>
#include <stdlib.h>
#include <os2.h>

#define INCL_OS2MM
#define INCL_MCIOS2
#define INCL_MMIOOS2

#include <os2me.h>

#include "Basic.hpp"
#include "Class_CDROM.hpp"
#include "Resource.hpp"

//---------------------------------------------------------
// Global constants

// The MSECTOMM macro generates an error in C++
const MMTIME MM100 = MSECTOMM(100);// 100 msec in MMTIME
const MMTIME MM1000 = MSECTOMM(1000);// 1 sec in MMTIME

//---------------------------------------------------------

void Class_CDROM::Set_MM() {
     // Set the default time to MMTIME
     // Not really necessary since the default format
     // is supposed to be MMTIME.  (Better be safe than
     // sorry)
     mciSetParameters.ulTimeFormat = MCI_FORMAT_MMTIME;
     ulrc = mciSendCommand(usDeviceID,
                           MCI_SET,
                           MCI_WAIT | MCI_SET_TIME_FORMAT,
                           &mciSetParameters,
                           0);
     if(ProcessReturnCode(ulrc, "MCI_SET_TIME_FORMAT")) {
          ErrorBox(message);
     }
}// Set_MM()

//---------------------------------------------------------

BOOL Class_CDROM::Open() {
     // Open the device for non-exclusive use

     // We're requesting notification since this action
     // takes a while.
     // Also, the first parameter is 0 because we do not
     // know the ID of the device.  However, it will be
     // immediately returned by the command.
     // Finally, we do not want to hog the device, so we
     // open in shareable mode.  However, this flag is not
     // sufficient to have an application that shares the
     // device.  WM_ACTIVATE and MM_MCIPASSDEVICE messages
     // are involved

     // This is the name of the device to be opened
     mciOpenParameters.pszDeviceType = MCI_DEVTYPE_CD_AUDIO_NAME;

     ulrc = mciSendCommand(0,
                           MCI_OPEN,
                           MCI_NOTIFY | MCI_OPEN_SHAREABLE,
                           &mciOpenParameters,
                           0);
     if(!ProcessReturnCode(ulrc, "MCI_OPEN")) {
          ErrorBox(message);
          return FALSE;
     }
     usDeviceID = mciOpenParameters.usDeviceID;
     return TRUE;
}// Open()

//---------------------------------------------------------

void Class_CDROM::Acquire() {
     // Acquire the device
     ulrc = mciSendCommand(usDeviceID,
                           MCI_ACQUIREDEVICE,
                           MCI_WAIT,
                           &mciGenericParameters,
                           NULL);
     if(!ProcessReturnCode(ulrc, "MCI_ACQUIREDEVICE")) {
          ErrorBox(message);
     }
}// Acquire()

//---------------------------------------------------------

void Class_CDROM::Close() {
     // Close the audio CD device

     // Let's stop the device first
     if(playing)
          Stop();

     ulrc = mciSendCommand(usDeviceID,
                           MCI_CLOSE,
                           MCI_WAIT,
                           &mciGenericParameters,
                           0);
     if(!ProcessReturnCode(ulrc, "MCI_CLOSE")) {
          ErrorBox(message);
     }
}// Close()

//---------------------------------------------------------

void Class_CDROM::Play() {
     // the mciPlayParameters.ulFrom field is set in
     // SetPlayPosition() or when Seeking()
     ulrc = mciSendCommand(usDeviceID,
                           MCI_PLAY,
                           MCI_NOTIFY | MCI_FROM,
                           &mciPlayParameters,
                           0);
     if(ProcessReturnCode(ulrc, "MCI_PLAY")) {
          EnablePlay(FALSE);
          EnablePause(TRUE);
          ActivatePause(TRUE);
     }
     else {
          ErrorBox(message);
     }
}// Play()

//---------------------------------------------------------

void Class_CDROM::Stop() {
     ulrc = mciSendCommand(usDeviceID,
                           MCI_STOP,
                           MCI_WAIT,
                           &mciGenericParameters,
                           0);
     if(ProcessReturnCode(ulrc, "MCI_STOP")) {
          EnablePlay(TRUE);
          EnablePause(TRUE);
          ActivatePause(FALSE);
          // Once we've stopped, the next play position
          // should be at the beginning of the track, or
          // possibly at the beginning of the CD.  But it
          // should definitively not be from the current
          // position since that's the behavior expected
          // from pause and resume.
          SetPlayPosition();
     }
     else {
          ErrorBox(message);
     }
}// Stop()

//---------------------------------------------------------

void Class_CDROM::Pause() {
     // Pause the CD device
     ulrc = mciSendCommand(usDeviceID,
                           MCI_PAUSE,
                           MCI_WAIT,
                           &mciGenericParameters,
                           0);
     if(ProcessReturnCode(ulrc, "MCI_PAUSE")) {
          EnablePause(FALSE);
     }
     else {
          ErrorBox(message);
     }
}// Pause()

//---------------------------------------------------------

void Class_CDROM::Resume() {
     // Resume playback
     ulrc = mciSendCommand(usDeviceID,
                           MCI_RESUME,
                           MCI_WAIT,
                           &mciGenericParameters,
                           0);
     if(ProcessReturnCode(ulrc, "MCI_RESUME")) {
          EnablePause(TRUE);
     }
     else {
          ErrorBox(message);
     }
}// Resume()

//---------------------------------------------------------

void Class_CDROM::Previous() {
     // Go to the previous track

     // First we determine the previous track
     PreviousTrack();

     // Then we prepare to play the next track
     SetPlayPosition();

     // And finally restart the current action
     RestartCurrentAction();
}// Previous()

//---------------------------------------------------------

void Class_CDROM::Next() {
     // Go to the next track

     // First we determine the next track,
     NextTrack();

     // Then we prepare to play the next track
     SetPlayPosition();

     // And finally restart the current action
     RestartCurrentAction();
}// Next()

//---------------------------------------------------------

void Class_CDROM::Insert() {
     // Close the tray door
     ulrc = mciSendCommand(usDeviceID,
                           MCI_SET,
                           MCI_WAIT | MCI_SET_DOOR_CLOSED,
                           &mciSetParameters,
                           0);
     if(ProcessReturnCode(ulrc, "MCI_INSERT")) {
          EnableEject(TRUE);
          // Delay so that the hardware will be ready when
          // we query it.  If DosSleep is to be used, then it
          // should be placed in a separate thread so as not
          // to block the message queue
          //DosSleep(7000);
          DetectCD();
     }
     else {
          ErrorBox(message);
     }
}// Insert()

//---------------------------------------------------------

void Class_CDROM::Eject() {
     // Open the tray door

     // First, we should remove the cuepoint
     RemoveCuePoint();

     // And then we eject
     ulrc = mciSendCommand(usDeviceID,
                           MCI_SET,
                           MCI_WAIT | MCI_SET_DOOR_OPEN,
                           &mciSetParameters,
                           0);
     if(ProcessReturnCode(ulrc, "MCI_EJECT")) {
          current_track = 0;
          mmtime = 0;
          WinPostMsg(hwnd, WM_USER_SHOW_TRACK, 0, 0);
          WinPostMsg(hwnd, WM_USER_SHOW_TIME, 0, 0);
          EnableEject(FALSE);
          DetectCD();
     }
     else {
          ErrorBox(message);
     }
}// Eject()

//---------------------------------------------------------

void Class_CDROM::Seeking(BOOL state) {
     // Start or stop seeking

     if(state) { // We've started seeking
          // Let's accurately determine the current
          // time location
          mciStatusParameters.ulItem = MCI_STATUS_POSITION;
          ulrc = mciSendCommand(usDeviceID,
                                MCI_STATUS,
                                MCI_WAIT | MCI_STATUS_ITEM,
                                &mciStatusParameters,
                                0);
          if(ProcessReturnCode(ulrc, "MCI_STATUS_POSITION in Seeking")) {
               // Update the current time
               mmtime = mciStatusParameters.ulReturn;
          }
          else {
               ErrorBox(message);
               return;
          }
          if(playing && !paused) {
               // We're stopping Play since we do not want
               // the CD to play while we're seeking
               ulrc = mciSendCommand(usDeviceID,
                                     MCI_STOP,
                                     MCI_WAIT,
                                     &mciGenericParameters,
                                     0);
               if(!ProcessReturnCode(ulrc, "MCI_STOP in Seeking")) {
                    ErrorBox(message);
                    return;
               }
          }
          // We'll fake actual seeking.  We start a timer
          // and everytime we receive a WM_TIMER message,
          // we'll update the display.
          if(!WinStartTimer(hab, hwnd, timerID, 150)) {
              ErrorBox("Error starting the timer in Seeking");
              return;
          }
     } // We've started seeking
     else { // We've stopped seeking
          // Stop the timer
          WinStopTimer(hab, hwnd, timerID);

          // Set the next play position to be from the
          // current time location
          mciPlayParameters.ulFrom = mmtime;

          // Go to that location
          SeekTo(mmtime);

          // And restart the current action
          RestartCurrentAction();
     } // We've stopped seeking
}// Seeking()

//---------------------------------------------------------

void Class_CDROM::Seek() {
     // Seek in the given direction
     if(SeekBackward) { // seek in the backward direction
          if((mmtime - MM1000) < TOC.record[current_track].ulStartAddr) {
               // We're at the beginning of the track

               // Let's seek to the beginning of the track
               // so that the MCI_STATUS_POSITION 'message'
               // will report that the hardware time
               // position is the beginning of the track
               SeekTo(TOC.record[current_track].ulStartAddr);

               // Move to the previous track
               PreviousTrack();
               mmtime = TOC.record[current_track].ulEndAddr;
          }
          else {
               mmtime -= MM1000;
          }
     } // seek in the backward direction
     else { // seek in the forward direction
          if((mmtime + MM1000) > TOC.record[current_track].ulEndAddr) {
               // We've passed the end of the track
               NextTrack();
               mmtime = TOC.record[current_track].ulStartAddr;
          }
          else {
               mmtime += MM1000;
          }
     } // seek in the forward direction

     WinSendMsg(hwnd, WM_USER_SHOW_TIME, 0, 0);
}// Seek()

//---------------------------------------------------------

/*
     // You can use this code to query the current volume level
     // of the device.  This is not necessary since we're keeping
     // track of this in the volume structure

     mciStatusParameters.ulItem = MCI_STATUS_VOLUME;
     ulrc = mciSendCommand(usDeviceID,
                           MCI_STATUS,
                           MCI_WAIT | MCI_STATUS_ITEM,
                           &mciStatusParameters,
                           0);
     if(ProcessReturnCode(ulrc, "MCI_STATUS_VOLUME")) {
          volume.left = LOUSHORT(mciStatusParameters.ulReturn);
          volume.right = HIUSHORT(mciStatusParameters.ulReturn);
     }
     else {
          ErrorBox(message);
     }
*/

//---------------------------------------------------------

void Class_CDROM::Set_Volume(Struct_Volume _volume) {
     // Set the volume for the left and right channels
     volume = _volume;

     // If the current device has not been specified
     // use the CD device (see Set_Connector)
     if(!VolumeDeviceID)
          VolumeDeviceID = usDeviceID;

     // specify the channels to modify and their values
     mciSetParameters.ulAudio = MCI_SET_AUDIO_ALL;
     mciSetParameters.ulLevel = (ULONG)MPFROM2SHORT(volume.left, volume.right);

     ulrc = mciSendCommand(VolumeDeviceID,
                           MCI_SET,
                           MCI_WAIT | MCI_SET_AUDIO |
                           MCI_SET_VOLUME | MCI_SET_ON,
                           &mciSetParameters,
                           NULL);
     if(!ProcessReturnCode(ulrc, "MCI_SET_VOLUME")) {
          ErrorBox(message);
     }
}// Set_Volume()

//---------------------------------------------------------

void Class_CDROM::Mute() {
     // Mute the cd-rom

     // You could specify individual channels, to mute
     // only the left or the right channel
     mciSetParameters.ulAudio = MCI_SET_AUDIO_ALL;

     // If the current device has not been specified
     // use the CD device (see Set_Connector)
     if(!VolumeDeviceID)
          VolumeDeviceID = usDeviceID;

     ulrc = mciSendCommand(VolumeDeviceID,
                           MCI_SET,
                           MCI_WAIT | MCI_SET_AUDIO | MCI_SET_OFF,
                           &mciSetParameters,
                           0);
     if(ProcessReturnCode(ulrc, "MCI_MUTE")) {
          EnableMute(FALSE);
          EnableMenuitem(IDM_LEVEL, FALSE);
     }
     else {
          ErrorBox(message);
     }
}// Mute()

//---------------------------------------------------------

void Class_CDROM::Unmute() {
     // Unmute the cd-rom

     // You could specify individual channels, to unmute
     // only the left or the right channel
     mciSetParameters.ulAudio = MCI_SET_AUDIO_ALL;

     // If the current device has not been specified
     // use the CD device (see Set_Connector)
     if(!VolumeDeviceID)
          VolumeDeviceID = usDeviceID;

     ulrc = mciSendCommand(VolumeDeviceID,
                           MCI_SET,
                           MCI_WAIT | MCI_SET_AUDIO | MCI_SET_ON,
                           &mciSetParameters,
                           0);
     if(ProcessReturnCode(ulrc, "MCI_UNMUTE")) {
          EnableMute(TRUE);
          EnableMenuitem(IDM_LEVEL, TRUE);
     }
     else {
          ErrorBox(message);
     }
}// Unmute()

//---------------------------------------------------------

void Class_CDROM::DetectCD() {
     // Detects whether an audio CD is present
     BOOL present;

     mciStatusParameters.ulItem = MCI_STATUS_READY;
     ulrc = mciSendCommand(usDeviceID,
                           MCI_STATUS,
                           MCI_WAIT | MCI_STATUS_ITEM,
                           &mciStatusParameters,
                           0);
     if(ProcessReturnCode(ulrc, "MCI_STATUS_READY")) {
          // Returns MCI_TRUE if the device is ready
          // otherwise MCI_FALSE is returned
          present = (mciStatusParameters.ulReturn == MCI_TRUE);
     }
     else {
          ErrorBox(message);
          present = FALSE;
     }

     // show these buttons by default
     EnablePlay(TRUE);
     EnablePause(TRUE);
     EnableMute(TRUE);

     // Play needs to be pressed to activate Pause
     ActivatePause(FALSE);

     // Activate/Deactivate these actions, depending on
     // whether there's an audio CD in the CD device
     ActivatePlay(present);
     ActivatePrevious(present);
     ActivateNext(present);
     ActivateBackward(present);
     ActivateForward(present);
     ActivateMute(present);
     EnableMenuitem(IDM_TOC, present);

     if(present) {
          // Obtain this CD's TOC
          Get_TOC();

          // This has to be done everytime we start a new CD
          SetPositionAdvise();

          // Reset current track and time
          current_track = 0;
          SetPlayPosition();
          SetCuePoint();
          WinSendMsg(hwnd, WM_USER_SHOW_TRACK, 0, 0);

          // Simply by issuing the Play() command in this
          // function, we're implementing AutoPlay:  as
          // soon as a media is detecting, we start playing
          // it.  The problem that is still unresolved is how
          // to know that we have to check for the presence
          // of a CD.  Querying the device periodically, with
          // the help of a timer would work.  Unfortunately,
          // polling a device is not a very good idea.
          Play();
     }
}// DetectCD()

//---------------------------------------------------------

void Class_CDROM::Get_Capacities() {
     // Query all of the device's capacities, one after the
     // other.  This could be done is a separate thread or
     // using MCI_NOTIFY.

     // Although we query all those capacities, the program
     // does not take notice of the results, except in the
     // connector dialog and for MCI_PLAY.

     mciGetDevCapsParameters.ulItem = MCI_GETDEVCAPS_CAN_EJECT;
     ulrc = mciSendCommand(usDeviceID,
                           MCI_GETDEVCAPS,
                           MCI_WAIT | MCI_GETDEVCAPS_ITEM,
                           &mciGetDevCapsParameters,
                           0);
     if(ProcessReturnCode(ulrc, "MCI_GETDEVCAPS_CAN_EJECT"))
          capacities.Eject = (mciGetDevCapsParameters.ulReturn == MCI_TRUE);
     else {
          ErrorBox(message);
          capacities.Eject = FALSE;
     }

     mciGetDevCapsParameters.ulItem = MCI_GETDEVCAPS_CAN_LOCKEJECT;
     ulrc = mciSendCommand(usDeviceID,
                           MCI_GETDEVCAPS,
                           MCI_WAIT | MCI_GETDEVCAPS_ITEM,
                           &mciGetDevCapsParameters,
                           0);
     if(ProcessReturnCode(ulrc, "MCI_GETDEVCAPS_CAN_LOCKEJECT"))
               capacities.LockEject = (mciGetDevCapsParameters.ulReturn == MCI_TRUE);
     else {
          ErrorBox(message);
          capacities.LockEject = FALSE;
     }

     mciGetDevCapsParameters.ulItem = MCI_GETDEVCAPS_CAN_PLAY;
     ulrc = mciSendCommand(usDeviceID,
                           MCI_GETDEVCAPS,
                           MCI_WAIT | MCI_GETDEVCAPS_ITEM,
                           &mciGetDevCapsParameters,
                           0);
     if(ProcessReturnCode(ulrc, "MCI_GETDEVCAPS_CAN_PLAY"))
          capacities.Play = (mciGetDevCapsParameters.ulReturn == MCI_TRUE);
     else {
          ErrorBox(message);
          capacities.Play = FALSE;
     }
     if(!capacities.Play) {
          ErrorBox("Critical error!\nThe device cannont perform basic tasks like PLAY");
          WinSendMsg(hwnd, WM_CLOSE, 0, 0);
     }

     mciGetDevCapsParameters.ulItem = MCI_GETDEVCAPS_CAN_PROCESS_INTERNAL;
     ulrc = mciSendCommand(usDeviceID,
                           MCI_GETDEVCAPS,
                           MCI_WAIT | MCI_GETDEVCAPS_ITEM,
                           &mciGetDevCapsParameters,
                           0);
     if(ProcessReturnCode(ulrc, "MCI_GETDEVCAPS_CAN_PROCESS_INTERNAL"))
          capacities.ProcessInternal = (mciGetDevCapsParameters.ulReturn == MCI_TRUE);
     else {
          ErrorBox(message);
          capacities.ProcessInternal = FALSE;
     }

     mciGetDevCapsParameters.ulItem = MCI_GETDEVCAPS_CAN_RECORD;
     ulrc = mciSendCommand(usDeviceID,
                           MCI_GETDEVCAPS,
                           MCI_WAIT | MCI_GETDEVCAPS_ITEM,
                           &mciGetDevCapsParameters,
                           0);
     if(ProcessReturnCode(ulrc, "MCI_GETDEVCAPS_CAN_RECORD"))
          capacities.Record = (mciGetDevCapsParameters.ulReturn == MCI_TRUE);
     else {
          ErrorBox(message);
          capacities.Record = FALSE;
     }

     mciGetDevCapsParameters.ulItem = MCI_GETDEVCAPS_CAN_RECORD_INSERT;
     ulrc = mciSendCommand(usDeviceID,
                           MCI_GETDEVCAPS,
                           MCI_WAIT | MCI_GETDEVCAPS_ITEM,
                           &mciGetDevCapsParameters,
                           0);
     if(ProcessReturnCode(ulrc, "MCI_GETDEVCAPS_CAN_RECORD_INSERT"))
          capacities.RecordInsert = (mciGetDevCapsParameters.ulReturn == MCI_TRUE);
     else {
          ErrorBox(message);
          capacities.RecordInsert = FALSE;
     }

     mciGetDevCapsParameters.ulItem = MCI_GETDEVCAPS_CAN_SAVE;
     ulrc = mciSendCommand(usDeviceID,
                           MCI_GETDEVCAPS,
                           MCI_WAIT | MCI_GETDEVCAPS_ITEM,
                           &mciGetDevCapsParameters,
                           0);
     if(ProcessReturnCode(ulrc, "MCI_GETDEVCAPS_CAN_SAVE"))
          capacities.Save = (mciGetDevCapsParameters.ulReturn == MCI_TRUE);
     else {
          ErrorBox(message);
          capacities.Save = FALSE;
     }

     mciGetDevCapsParameters.ulItem = MCI_GETDEVCAPS_CAN_SETVOLUME;
     ulrc = mciSendCommand(usDeviceID,
                           MCI_GETDEVCAPS,
                           MCI_WAIT | MCI_GETDEVCAPS_ITEM,
                           &mciGetDevCapsParameters,
                           0);
     if(ProcessReturnCode(ulrc, "MCI_GETDEVCAPS_CAN_SETVOLUME"))
          capacities.SetVolume = (mciGetDevCapsParameters.ulReturn == MCI_TRUE);
     else {
          ErrorBox(message);
          capacities.SetVolume = FALSE;
     }

     // In the case of the CD player, it's better to rely
     // on the connectors that are detected rather than on
     // this query
     mciGetDevCapsParameters.ulItem = MCI_GETDEVCAPS_CAN_STREAM;
     ulrc = mciSendCommand(usDeviceID,
                           MCI_GETDEVCAPS,
                           MCI_WAIT | MCI_GETDEVCAPS_ITEM,
                           &mciGetDevCapsParameters,
                           0);
     if(ProcessReturnCode(ulrc, "MCI_GETDEVCAPS_CAN_STREAM"))
          capacities.StreamDigital = (mciGetDevCapsParameters.ulReturn == MCI_TRUE);
     else {
          ErrorBox(message);
          capacities.StreamDigital = FALSE;
     }

     mciGetDevCapsParameters.ulItem = MCI_GETDEVCAPS_HAS_AUDIO;
     ulrc = mciSendCommand(usDeviceID,
                           MCI_GETDEVCAPS,
                           MCI_WAIT | MCI_GETDEVCAPS_ITEM,
                           &mciGetDevCapsParameters,
                           0);
     if(ProcessReturnCode(ulrc, "MCI_GETDEVCAPS_HAS_AUDIO"))
          capacities.Audio = (mciGetDevCapsParameters.ulReturn == MCI_TRUE);
     else {
          ErrorBox(message);
          capacities.Audio = FALSE;
     }

     // This will generate an error, for some unknown reason.
     // MMREF states that "The following list of items can
     // be used regardless of the type of device:" and this
     // item is part of the list.
     mciGetDevCapsParameters.ulItem = MCI_GETDEVCAPS_HAS_IMAGE;
     ulrc = mciSendCommand(usDeviceID,
                           MCI_GETDEVCAPS,
                           MCI_WAIT | MCI_GETDEVCAPS_ITEM,
                           &mciGetDevCapsParameters,
                           0);
     if(ProcessReturnCode(ulrc, "MCI_GETDEVCAPS_HAS_IMAGE"))
          capacities.Image = (mciGetDevCapsParameters.ulReturn == MCI_TRUE);
     else {
          ErrorBox(message);
          capacities.Image = FALSE;
     }

     mciGetDevCapsParameters.ulItem = MCI_GETDEVCAPS_HAS_VIDEO;
     ulrc = mciSendCommand(usDeviceID,
                           MCI_GETDEVCAPS,
                           MCI_WAIT | MCI_GETDEVCAPS_ITEM,
                           &mciGetDevCapsParameters,
                           0);
     if(ProcessReturnCode(ulrc, "MCI_GETDEVCAPS_HAS_VIDEO"))
          capacities.Video = (mciGetDevCapsParameters.ulReturn == MCI_TRUE);
     else {
          ErrorBox(message);
          capacities.Video = FALSE;
     }

     // Query the number of connectors of the specified type
     // According to MMPG, a CD audio supports two
     // connectors:  headphones and stream.  However,
     // mine does not support stream, so my code has not
     // been tested with that type of hardware
     mciConnectorInfoParameters.ulDeviceTypeID = MCI_DEVTYPE_CD_AUDIO;
     mciConnectorInfoParameters.ulConnectorType = MCI_HEADPHONES_CONNECTOR;
     ulrc = mciSendCommand(usDeviceID,
                           MCI_CONNECTORINFO,
                           MCI_ENUMERATE_CONNECTORS | MCI_WAIT | MCI_CONNECTOR_TYPE,
                           &mciConnectorInfoParameters,
                           0);
     if(ProcessReturnCode(ulrc, "MCI_CONNECTOR_INFO headphones")) {
          capacities.Headphones = (int) mciConnectorInfoParameters.ulReturn;
     }
     else {
          ErrorBox(message);
          capacities.Headphones = 0;
     }

     mciConnectorInfoParameters.ulConnectorType = MCI_CD_STREAM_CONNECTOR;
     ulrc = mciSendCommand(usDeviceID,
                           MCI_CONNECTORINFO,
                           MCI_ENUMERATE_CONNECTORS | MCI_WAIT | MCI_CONNECTOR_TYPE,
                           &mciConnectorInfoParameters,
                           0);
     if(ProcessReturnCode(ulrc, "MCI_CONNECTOR_INFO CD stream")) {
          capacities.Stream = (int) mciConnectorInfoParameters.ulReturn;
     }
     else {
          ErrorBox(message);
          capacities.Stream = 0;
     }

/*
     // Enumerate all the valid connectors to the CD device
     // In theory, this should work.  In practice, I don't
     // know;  it only reports headphones.
{
     int ConnectorsCount;
     ulrc = mciSendCommand(0,
                         MCI_CONNECTORINFO,
                         MCI_ENUMERATE_CONNECTORS | MCI_WAIT,
                         &mciConnectorInfoParameters,
                         0);
     if(ProcessReturnCode(ulrc, "MCI_ENUMERATE_CONNECTORS")) {
          ConnectorsCount = (int) mciConnectorInfoParameters.ulReturn;
          for(int x = 1; x <= ConnectorsCount; x++) {
               mciConnectorInfoParameters.ulConnectorIndex = x;
               ulrc = mciSendCommand(0,
                                     MCI_CONNECTORINFO,
                                     MCI_QUERY_CONNECTOR_TYPE | MCI_CONNECTOR_INDEX |MCI_WAIT,
                                     &mciConnectorInfoParameters,
                                     0);
               if(ProcessReturnCode(ulrc, "MCI_QUERY_CONNECTOR_TYPE")) {
                    switch(mciConnectorInfoParameters.ulReturn) {
                    case MCI_CD_STREAM_CONNECTOR:
                         InfoBox("stream connector");
                         break;
                    case MCI_HEADPHONES_CONNECTOR:
                         InfoBox("headphones connector");
                         break;
                    default:
                         InfoBox("Other connector");
                         break;
                    }
               }
               else {
                    ErrorBox(message);
               }
          }
     }
     else {
          ErrorBox(message);
          ConnectorsCount = 0;
     }

}
*/

}// Get_Capacities()

//---------------------------------------------------------

void Class_CDROM::Get_TOC() {
     // Obtain the Table of Contents for the audio CD

     // Query the number of audio tracks
     mciStatusParameters.ulItem = MCI_STATUS_NUMBER_OF_TRACKS;
     ulrc = mciSendCommand(usDeviceID,
                           MCI_STATUS,
                           MCI_WAIT | MCI_STATUS_ITEM,
                           &mciStatusParameters,
                           NULL);
     if(ProcessReturnCode(ulrc, "MCI_STATUS_NUMBER_OF_TRACKS")) {
          TOC.track_count = last_track = (int) mciStatusParameters.ulReturn;
          // Release the memory for the previous TOC
          delete [] TOC.record;
          // Allocate memory for the new TOC
          TOC.record = new MCI_TOC_REC [TOC.track_count];
          // Set the structure data
          TOC.tocHeader.pBuf = TOC.record;
          TOC.tocHeader.ulBufSize = TOC.track_count * sizeof(MCI_TOC_REC);
          // Send the message
          ulrc = mciSendCommand(usDeviceID,
                                MCI_GETTOC,
                                MCI_WAIT,
                                &TOC.tocHeader,
                                NULL);
          if(ProcessReturnCode(ulrc, "MCI_GETTOC")) {
               // Select the first track
               current_track = 0;
               SetCuePoint();
               SetPlayPosition();
          }
          else {
               ErrorBox(message);
               return;
          }
     }
     else {
          ErrorBox(message);
          return;
     }
}// Get_TOC()

//---------------------------------------------------------

void Class_CDROM::TOCtoa(char *text) {
     // Convert the TOC structure into an ascii string

     int x;
     char temp[100];
     char start_time[100], end_time[100], length_time[100];

     // Set the header
     strcpy(text, "Track   Start        End       Length\n");

     // Add each entry
     for(x = 0; x < last_track; x++) {
          mmtoa(TOC.record[x].ulStartAddr, start_time);
          mmtoa(TOC.record[x].ulEndAddr, end_time);
          mmtoa(TOC.record[x].ulEndAddr - TOC.record[x].ulStartAddr, length_time);

          sprintf(temp,
                  "   %02d    %s - %s  (%s)\n",
                  TOC.record[x].TrackNum,
                  start_time,
                  end_time,
                  length_time);
          strcat(text, temp);
     }
}// TOCtoa()

//---------------------------------------------------------

void Class_CDROM::mmtoa(MMTIME mm, char *text) {
     // Convert MMTIME to an ascii string

     ULONG hms;// MMTIME in HourMinuteSecond format
     int hour, minute, second;// Individual values

     hour = minute = second = 0;
     strcpy(text, "");

     // Convert MMTIME to HMS
     hms = HMSFROMMM(mm);

     // Retrieve each individual value
     hour = int(ULONG_LWLB(hms));// The ULONG_LWLB macro enerates an error in C++
     minute = int(ULONG_LWHB(hms));
     second = int(ULONG_HWLB(hms));

     // Generate the ascii string
     sprintf(text, "%02d:%02d:%02d", hour, minute, second);
}// mmtoa()

//---------------------------------------------------------

MMTIME Class_CDROM::PositionInTrack() {
     // Where are we in the track?

     mciStatusParameters.ulItem = MCI_STATUS_POSITION_IN_TRACK;
     ulrc = mciSendCommand(usDeviceID,
                           MCI_STATUS,
                           MCI_WAIT | MCI_STATUS_ITEM,
                           &mciStatusParameters,
                           0);
     if(ProcessReturnCode(ulrc, "MCI_STATUS_POSITION_IN_TRACK")) {
          return mciStatusParameters.ulReturn;
     }
     else {
          ErrorBox(message);
          return 0;
     }
}// PositionInTrack()

//---------------------------------------------------------

MMTIME  Class_CDROM::Get_CorrectedTime(MMTIME time) {
     // Returns the time corrected for the specified
     // time format

     MMTIME Corrected;// Time to return

     switch(TimeFormat) {
     case IDM_TIME_TRACKPLAYED:
          // specified time - start time of track
          // (The audio CD might not start playing at time
          //  0:00)
          Corrected = time
                      - TOC.record[current_track].ulStartAddr;
          break;
     case IDM_TIME_TRACKREMAIN:
          // endtime of track - specified time
          Corrected = TOC.record[current_track].ulEndAddr
                      - time;
          break;
     case IDM_TIME_CDPLAYED:
          // specified time - starttime of first track
          // (The audio CD might not start playing at time
          //  0:00)
          Corrected = time
                       - TOC.record[0].ulStartAddr;
          break;
     case IDM_TIME_CDREMAIN:
          // end time of last track - specified time
          // - start time of first track
          // (The audio CD might not start playing at time
          //  0:00)
          Corrected = TOC.record[last_track - 1].ulEndAddr
                      - time
                      - TOC.record[0].ulStartAddr;
          break;
     default:
          // The programmer forgot to set the time format
          // when the window was created
          ErrorBox("Time Display was not properly set");
          break;
     }
     return Corrected;
}// Get_CorrectedTime()

//---------------------------------------------------------

void Class_CDROM::Set_Connector(int _connector) {
     // Set the connector

     // I haven't properly tested this code since my hardware
     // only supports one connector (headphones).

     // If the new selector is the same as the current one
     // do nothing
     if(connector == _connector)
          return;

     // Enable the new connector
     connector = _connector;

     mciConnectorParameters.ulConnectorType = connector;
     ulrc = mciSendCommand(usDeviceID,
                           MCI_CONNECTOR,
                           MCI_ENABLE_CONNECTOR | MCI_CONNECTOR_TYPE | MCI_WAIT,
                           &mciConnectorParameters,
                           0);
     if(ProcessReturnCode(ulrc, "MCI_ENABLE_CONNECTOR")) {
          // If we change the connector (to stream or
          // headphones), then we have to obtain the ID
          // of the connected device (CD or AmpMix) in
          // order to be able to set the volume.  We can
          // obtain this ID from the mciConnection
          // structure

          // For some reason, this message generates an
          // error message the first time it is used.
          mciConnectionParameters.ulConnectorType = connector;
          ulrc = mciSendCommand(usDeviceID,
                                MCI_CONNECTION,
                                MCI_QUERY_CONNECTION | MCI_WAIT,
                                &mciConnectionParameters,
                                0);
          if(ProcessReturnCode(ulrc, "MCI_QUERY_CONNECTION")) {
               VolumeDeviceID = mciConnectionParameters.usToDeviceID;
               // This may not be necessary
               // If it is, and you want to restart Play
               // from the current location, you'd have to
               // query the current hardware time position
               // (MCI_STATUS_POSITION) and set
               // mciPlayParameters.ulFrom to that hardware
               // time location before issuing:
               //RestartCurrentAction();
               // You may also have to reenable the
               // notifications for position changes with:
               // SetPositionAdvise();
          }
          else {
               ErrorBox(message);

               // We'll use the ID of the CD device
               VolumeDeviceID = usDeviceID;
          }
     }
     else {
          ErrorBox(message);
     }
}// Set_Connector()

//---------------------------------------------------------

void Class_CDROM::SeekTo(MMTIME time) {
     // Seek the device to the specified time

     mciSeekParameters.ulTo = time;
     ulrc = mciSendCommand(usDeviceID,
                           MCI_SEEK,
                           MCI_WAIT | MCI_TO,
                           &mciSeekParameters,
                           0);
     if(!ProcessReturnCode(ulrc, "MCI_SEEK")) {
          ErrorBox(message);
     }
}// SeekTo()

//---------------------------------------------------------

void Class_CDROM::SetPlayPosition() {
     // Set the next play position to occur at the
     // beginning of the current track
     mciPlayParameters.ulFrom = TOC.record[current_track].ulStartAddr;

     // We seek to that track if we're not playing
     // If we were playing, the next Play command will
     // be from the location we've just specified above
     if(paused || !playing) {
          SeekTo(TOC.record[current_track].ulStartAddr);
     }

     // Finally, reset the current time
     mmtime = TOC.record[current_track].ulStartAddr;
     WinSendMsg(hwnd,WM_USER_SHOW_TIME, 0, 0);
}// SetPlayPosition()

//---------------------------------------------------------

void Class_CDROM::SetPositionAdvise() {
          // Request a notification every 1 sec of play
          // so we can update the time display

          mciPositionParameters.ulUnits = MM1000;
          ulrc = mciSendCommand(usDeviceID,
                                MCI_SET_POSITION_ADVISE,
                                MCI_WAIT | MCI_SET_POSITION_ADVISE_ON,
                                &mciPositionParameters,
                                0);
          if(!ProcessReturnCode(ulrc, "MCI_SET_POSITION_ADVISE")) {
               ErrorBox(message);
          }
}// SetPositionAdvise()

//---------------------------------------------------------

void Class_CDROM::RestartCurrentAction() {
     // Restarts the previous action.  If we were
     // paused, we first have to Play and then
     // Pause, so that if Resume is selected,
     // we'll resume from the correct time location
     if(playing) {
          if (PositionInTrack() > MM100) {
               ulrc = mciSendCommand(usDeviceID,
                                     MCI_PLAY,
                                     MCI_NOTIFY | MCI_FROM,
                                     &mciPlayParameters,
                                     0);
               if(!ProcessReturnCode(ulrc, "MCI_PLAY in RestartCurrentAction")) {
                    ErrorBox(message);
               }
          }
     }
     if(paused) {
          ulrc = mciSendCommand(usDeviceID,
                                MCI_PAUSE,
                                MCI_WAIT,
                                &mciGenericParameters,
                                0);
          if(!ProcessReturnCode(ulrc, "MCI_PAUSE in RestartCurrentAction")) {
               ErrorBox(message);
          }
     }

}// RestartCurrentAction()

//---------------------------------------------------------

void Class_CDROM::RemoveCuePoint() {
     // Removes the current cuepoint

     if(old_cue_point) {
          mciCuepointParameters.ulCuepoint = old_cue_point;
          ulrc = mciSendCommand(usDeviceID,
                                MCI_SET_CUEPOINT,
                                MCI_WAIT | MCI_SET_CUEPOINT_OFF,
                                &mciCuepointParameters,
                                0);
          if(!ProcessReturnCode(ulrc, "MCI_SET_CUEPOINT (Remove)")) {
               ErrorBox(message);
          }
          old_cue_point = 0;
     }
}// RemoveCuePoint()

//---------------------------------------------------------

void Class_CDROM::SetCuePoint() {
     // A cuepoint is set at the end of the end of the
     // current track so that we can receive a
     // notification and act appropriately.
     // And alternate approach would be to test the
     // current location when we receive the
     // MM_MCIPOSITIONCHANGE message.

     // First we remove the current cue point since there
     // can only be one cuepoint at a time
     RemoveCuePoint();

     // Then we set the new cue point
     mciCuepointParameters.ulCuepoint = TOC.record[current_track].ulEndAddr;
     ulrc = mciSendCommand(usDeviceID,
                           MCI_SET_CUEPOINT,
                           MCI_WAIT | MCI_SET_CUEPOINT_ON,
                           &mciCuepointParameters,
                           0);
     if(!ProcessReturnCode(ulrc, "MCI_SET_CUEPOINT (new)")) {
          ErrorBox(message);
     }

     // Finally, we set the old cue point to the one
     // we've just set, since that's the one that will
     // have to be removed the next time we want to set
     // a new cuepoint
     old_cue_point = TOC.record[current_track].ulEndAddr;
}// SetCuePoint()

//---------------------------------------------------------


